/*
 * GLX Hardware Device Driver for Matrox G200/G400
 * Copyright (C) 1999 Jeff Hartmann
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included
 * in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
 * JEFF HARTMANN, OR ANY OTHER CONTRIBUTORS BE LIABLE FOR ANY CLAIM, 
 * DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR 
 * OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE 
 * OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 *
 *
 * original by Jeff Hartmann <slicer@ionet.net>
 * 6/16/99: rewrite by John Carmack <johnc@idsoftware.com>
 */

#include <stdlib.h>
#include <errno.h>
#include <unistd.h>
#include <sys/mman.h>
#include <stdio.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <signal.h>

/* pass --enable-agp to configure instead */
/* #define USE_GART */

#ifdef USE_GART
#include <linux/agpgart.h>
#endif

#include "context.h"
#include "depth.h"
#include "macros.h"
#include "texstate.h"
#include "triangle.h"
#include "vb.h"
#include "types.h"

#include "xsmesaP.h"
#include "glx_log.h"
#include "mesaglx/context.h"
#include "mesaglx/matrix.h"
#include "mesaglx/types.h"

#define GC XXGC
#include "gcstruct.h"
#include "pixmapstr.h"
#include "servermd.h" /* PixmapBytePad */
#include "scrnintstr.h"
#include "regionstr.h"
#include "windowstr.h"
#undef GC

#include "vga.h"
#include "vgaPCI.h"
#include "mga.h"
#include "xaa/xf86xaa.h"

#include "mm.h"
#include "mgabuf.h"
#include "mgadd.h"
#include "g200_mac.h"
#include "mgalib.h"
#include "mgalog.h"


#ifndef NO_MTRR

#define MTRR_NEED_STRINGS
#include <errno.h>
#include <asm/mtrr.h>
#include <sys/ioctl.h>

static int mtrr;

#endif

/*

GLX_MGA_DMA=0	: virtual buffer, pseudo dma
GLX_MGA_DMA=1	: physical buffer, pseudo dma
GLX_MGA_DMA=2	: physical buffer, real dma
GLX_MGA_DMA=3	: physical buffer, real dma, async

GLX_MGA_DMAADR=92	: put the physical buffer at offset 92 * 0x100000 
GLX_MGA_DMASIZE=4	: use 4 * 0x100000 bytes of memory for the buffers

GLX_MGA_CMDSIZE=4	: use 4 * 0x100000 bytes for commands, the rest for textures

GLX_MGA_CARDCMDS=1	: put command buffers on card memory ( DOESN'T WORK! )
GLX_MGA_SYSTEMTEXTURE = 1	: put textures in main pci/agp memory instead of on card

*/

/* public vars */
mgaUI32		mgaDmaBufferDwords;		/* for overflow checks, forces flush */
mgaDma_buffer	*dma_buffer;			/* dmaBuffers[ activeDmaBuffer ] */

memHeap_t	*cardHeap;
mgaUI32		cardPhysical;
unsigned char	*cardVirtual;

memHeap_t	*textureHeap;
mgaUI32		textureHeapPhysical;		/* 0 if we aren't using PCI texturing */
unsigned char *textureHeapVirtual;			/* correct for either local or PCI heaps */

memHeap_t	*sysmemHeap;				/* physical memory, if available */
mgaUI32		sysmemBytes;				/* size of memory block */
mgaUI32		sysmemPhysical;			/* 0 if we don't have a physical mapping */
unsigned char	*sysmemVirtual;

int use_agp; /* will be set to "PDEA_pagpxfer_enable" if enabled or 0 if not */

/* private vars */
static	mgaDma_buffer	*dmaBuffers[2];
static	mgaUI32	activeDmaBuffer;		/* toggles between 0 and 1 */

static	mgaUI32	pseudoDmaPhysical;		/* physical address of pdma window */
static	mgaUI32	*pseudoDmaVirtual;		/* our local mapping of pdma window */

static	mgaUI32	bufferBytes;			/* size of buffer */
static	mgaUI32	bufferPhysical;		/* 0 if we don't have a physical mapping */
static	unsigned char	*bufferVirtual;
static	mgaUI32	dmaEnd;				/* value put in MGAREG_PRIMEND */

static	mgaUI32	largestDmaBuffer;		/* track flush sizes */

static	mgaUI32	registersLocked;		/* true when the registers are memory protected */

/*
Before we submit a dma command buffer, we write SYNC_DMA_BUSY to DWGSYNC.
The last command in the dma buffer will clear DWGSYNC.
The X server might also change the value in DWGSYNC, but the register
locking guarantees that dma will complete first.  We will spin until
timeout in waitDmaCompletion() if the X server ever happens to use
our magic value.

Any time DWGSYNC is read and it holds SYNC_DMA_BUSY, then dma is still
active and we need to wait until it is cleared.
*/
#define SYNC_DMA_BUSY	0xea832534		/* just a random number */


#define AGP_MODE_1   0x00000001
#define AGP_MODE_2   0x00000002
#define AGP_ENABLE   0x00000100
#define AGP_SBA      0x00000200
#define AGP_RQDEPTH  0x1f000000

#ifdef USE_GART
int    gartfd;
void   *gartbuf;
struct gart_info gartinf;
#endif

static void UnlockRegisters( void );

static int	usec( void ) {
	struct timeval tv;
	struct timezone tz;
	
	gettimeofday( &tv, &tz );
	
	return (tv.tv_sec & 2047) * 1000000 + tv.tv_usec;
}

static void delay( void ) {
}

static int waitForDmaCompletion( void ) {
	int		delays;
	int		primAdr;

	delays = 0;
	while ( 1 ) {
		primAdr = INREG( MGAREG_DWGSYNC );
		if ( primAdr != SYNC_DMA_BUSY ) {
			mgaMsg(10,"wait completed\n" );
			break;
		}
		mgaMsg(10,"wait\n" );

#if 1
		/* this is the system-friendly delay */
		usleep( 1000 );
		if ( delays > 300 ) {
			mgaMsg( 1, "waitForDmaCompletion timed out (%x)\n", primAdr );
			break;
		}
#else
		/* this delay spins in place, and may give higher performance */
		{
			int		i;
			for ( i = 0 ; i < 1000 ; i++ ) {
				delay();
			}
		}
#endif
		delays++;
	}
	
	if ( registersLocked ) {
		UnlockRegisters();
	}

	return delays;
}


static void MapPseudoDmaWindow( void ) {
	pseudoDmaPhysical =  (pcibusRead(MGAPciTag, 0x00000018)) & 0xff800000;
	pseudoDmaVirtual = 
		 xf86MapVidMem(vga256InfoRec.scrnIndex, EXTENDED_REGION,
			   (pointer) ((unsigned long) pseudoDmaPhysical),
			   0x800000);
	mgaMsg( 1, "pseudoDmaPhysical : %p\n", pseudoDmaPhysical );
	mgaMsg( 1, "pseudoDmaVirtual : %p\n", pseudoDmaVirtual );
}

static const char *getenvSafe(const char *name ) {
	const char	*r;
	
	r = getenv( name );
	if ( !r ) {
		return "";
	}
	return r;
}


static void MgaCloseMTRR( void ) {
   #ifndef NO_MTRR
	close(mtrr);
   #endif
}

static void MgaOpenMTRR( void ) {
#ifndef NO_MTRR
	if ((mtrr = open("/proc/mtrr", O_WRONLY, 0)) == -1) {
		if (errno == ENOENT) {
			  mgaError("/proc/mtrr not found: MTRR not enabled\n");
		}  else {
			  mgaError("Error opening /proc/mtrr: %s\n",strerror(errno));
			  mgaError("MTRR not enabled\n");
		}
		return;
	}
	atexit( MgaCloseMTRR );
#endif
}

/*
 * SetWriteCombining
 *
 * Checks for unsetting an existing MTRR if needed
 */
static void SetWriteCombining( long physical, int bytes ) {
#ifndef NO_MTRR
	struct mtrr_sentry sentry;
	struct mtrr_gentry gentry;
	int	i;
	
	if ( !mtrr ) {
		return;
	}
	/* remove any MTRR that conver the range */
	for ( i = 0 ; i < 128 ; i++ ) {
		gentry.regnum = i;
		if (ioctl(mtrr, MTRRIOC_GET_ENTRY, &gentry) == -1) {
			break;
		}
		mgaMsg(1,"MTRR reg %i: addr=0x%x size=0x%x type=%i\n", i, 
			gentry.base, gentry.size, gentry.type );
		if ( gentry.base >= physical + bytes ) {
			continue;
		}
		if ( gentry.base + gentry.size <= physical ) {
			continue;
		}
		
		/* we must delete this entry */
		sentry.base = gentry.base;
		sentry.size = gentry.size;
		if (ioctl(mtrr, MTRRIOC_DEL_ENTRY, &sentry) == -1) {
			mgaError("Error doing MTRRIOC_DEL_ENTRY on /proc/mtrr: %s\n", strerror(errno));
		} else {
			mgaMsg(1,"MTRRIOC_DEL_ENTRY succeeded\n" );
		}

		/* recreate fragments around the new region if necessary */
		if ( gentry.base < physical ) {
			sentry.base = gentry.base;
			sentry.size = physical - sentry.base;
			sentry.type = gentry.type;
			
			if ( ioctl(mtrr, MTRRIOC_ADD_ENTRY, &sentry) == -1 ) {
				mgaError("Error doing ioctl(2) on /proc/mtrr: %s\n", strerror(errno));
			} else {
				mgaMsg(1,"MTRR fragment added: addr=0x%x size=0x%x type=%i\n",
				   sentry.base, sentry.size, sentry.type );
			}
		}
		if ( gentry.base + gentry.size > physical + bytes ) {
			sentry.base = physical + bytes;
			sentry.size = gentry.base + gentry.size - sentry.base;
			sentry.type = sentry.type;
			
			if ( ioctl(mtrr, MTRRIOC_ADD_ENTRY, &sentry) == -1 ) {
				mgaError("Error doing ioctl(2) on /proc/mtrr: %s\n", strerror(errno));
			} else {
				mgaMsg(1,"MTRR fragment added: addr=0x%x size=0x%x type=%i\n",
				   sentry.base, sentry.size, sentry.type );
			}
		}
		
		/* because we deleted an entry, we need to check this index again */
		i--;
	}

	/* set this range to write combining */
	sentry.base = physical;
	sentry.size = bytes;
	sentry.type = 1; /* write-combining */

	if ( ioctl(mtrr, MTRRIOC_ADD_ENTRY, &sentry) == -1 ) {
		mgaError("Error doing ioctl(2) on /proc/mtrr: %s\n", strerror(errno));
		mgaError("MTRR not enabled\n");
	} else {
		mgaMsg(1,"MTRR enabled: write-combining, addr=0x%x size=0x%x\n",
			   sentry.base, sentry.size );
	}
	
#endif
}


static void MemoryBenchmark( void *buffer, int dwords ) {
	int		i;
	int		start, end;
	int		mb;
	int 		*base;
	
	base = (int *)buffer;
	
	start = usec();
	for ( i = 0 ; i < dwords ; i += 8 ) {
		base[i] = 
		base[i+1] =
		base[i+2] = 
		base[i+3] = 
		base[i+4] = 
		base[i+5] = 
		base[i+6] = 
		base[i+7] = 0x15151515;		/* dmapad nops */
	}
	end = usec();

	mb = ( (float)dwords / 0x40000 ) * 1000000 / (end - start);
		
	mgaMsg( 1, "MemoryBenchmark: %i mb/s\n", mb );
}

static void DmaBenchmark( unsigned int physical, int dwords ) {
#if 0	/* need to add the DWGSYNC at end of buffer */
	int		start, end;
	int		mb;
	float		fsec;

  	start = usec();

	dmaEnd = physical + dwords*4;

	OUTREG( MGAREG_PRIMADDRESS, physical );
	OUTREG( MGAREG_PRIMEND, dmaEnd | use_agp);

	waitForDmaCompletion();
	
	end = usec();

	fsec = ( end - start ) / 1000000.0;

	mb = ( (float)dwords * 4 / 0x100000 ) / fsec;

	mgaMsg( 1, "DmaBenchmark 0x%x bytes, %5.3f sec: %i mb/s\n", dwords*4, fsec, mb );a
#endif
}

static void AllocateCardDmaBuffer( void ) {
	PMemBlock	block;
	
	bufferBytes = 0x100000;
	block = mmAllocMem( cardHeap, bufferBytes, 8, 0 );
	if ( !block ) {
		mgaMsg( 1, "failed to allocate 0x%x bytes from cardHeap for command buffers.\n"
			, bufferBytes );
		return;
	}
	mgaMsg( 1, "allocated 0x%x bytes from cardHeap for command buffers.\n"
		, bufferBytes );
	bufferVirtual = cardVirtual + mmOffset( block );
	bufferPhysical = cardPhysical + mmOffset( block );
}

static void AllocatePhysicalDmaBuffer( void ) {
	PMemBlock	block;
	
	/* determine total size of buffer */
	bufferBytes = atoi( getenvSafe("GLX_MGA_CMDSIZE") );
	if ( !bufferBytes ) {
		mgaMsg(1,"defaulting to GLX_MGA_CMDSIZE = 4\n" );
		bufferBytes = 4;
	} else {
		mgaMsg(1,"using GLX_MGA_CMDSIZE = %i\n", bufferBytes );
	}
	bufferBytes *= 0x100000;
	
	block = mmAllocMem( sysmemHeap, bufferBytes, 8, 0 );
	if ( !block ) {
		mgaMsg( 1, "failed to allocate 0x%x bytes from sysmemHeap for command buffers.\n"
			, bufferBytes );
		return;
	}
	mgaMsg( 1, "allocated 0x%x bytes from sysmemHeap for command buffers.\n"
		, bufferBytes );
	bufferVirtual = sysmemVirtual + mmOffset( block );
	bufferPhysical = sysmemPhysical + mmOffset( block );
}	

static void AllocateVirtualDmaBuffer( void ) {
	/* determine total size of buffer */
	bufferBytes = atoi( getenvSafe("GLX_MGA_CMDSIZE") );
	if ( !bufferBytes ) {
		mgaMsg(1,"defaulting to GLX_MGA_CMDSIZE = 4\n" );
		bufferBytes = 4;
	} else {
		mgaMsg(1,"using GLX_MGA_CMDSIZE = %i\n", bufferBytes );
	}
	bufferBytes *= 0x100000;
	bufferVirtual = malloc( bufferBytes + 0x1000);
	/* align it to page size, might help on something used as much as this */
	bufferVirtual = (pointer)(((unsigned long) bufferVirtual & ~0xFFF) + 0x1000);
	mgaMsg( 1, "allocated 0x%x bytes from virtual memory for command buffers.\n"
		, bufferBytes );
}

static void OptimizeDMA( void ) {
	int		option;

	/* turn on enhmemacc */
	option = pcibusRead( MGAPciTag, 0x40 );
	mgaMsg(1,"option: 0x%x\n", option );		
	pcibusWrite(MGAPciTag, 0x00000040, option | (1<<22) );
}

#ifdef USE_GART  
static void OptimizeDMA2( void ) {

  	int mode = 1;
	/* turn on AGP 1x mode */
  	mgaMsg(1,"Enabling agp x1 mode\n");
	ioctl(gartfd, GARTIOCMODE, &mode);
  	OUTREG(MGAREG_AGP_PLL, AGP_PLL_agp2xpllen_disable);
	pcibusWrite(MGAPciTag, 0x000000f8, (AGP_MODE_1 | AGP_ENABLE | AGP_SBA | AGP_RQDEPTH));
}

static void OptimizeDMA3( void ) {
  	int mode = 2;
	/* turn on AGP 2x mode */
  	mgaMsg(1,"Enabling agp x2 mode\n");
	ioctl(gartfd, GARTIOCMODE, &mode);
  	OUTREG(MGAREG_AGP_PLL, AGP_PLL_agp2xpllen_enable);
	pcibusWrite(MGAPciTag, 0x000000f8, (AGP_MODE_2 | AGP_ENABLE | AGP_SBA | AGP_RQDEPTH));
}
#endif

/*
 * AllocateCommandBuffers
 * The dma command buffers can be either virtual or in the sysmemHeap
 * I never got card allocated buffers functioning reliably...
 */
static void AllocateCommandBuffers( void ) {
	/* try to allocate the command buffers in either sysmem or cardmem */
	if ( mgaglx.dmaDriver > 0 ) {
		if ( getenv("GLX_MGA_CARDCMDS") ) {
			AllocateCardDmaBuffer();
		} else if ( sysmemHeap ) {
			AllocatePhysicalDmaBuffer();
		}
	}
		
	/* if we didn't get real memory, get a virtual buffer and use PDMA */
	if ( !bufferPhysical ) {
		mgaglx.dmaDriver = 0;
		AllocateVirtualDmaBuffer();
	}
	
	/* benchmark the writing speed to the command buffer */
	MemoryBenchmark( bufferVirtual, bufferBytes / 4 );
	MemoryBenchmark( bufferVirtual, bufferBytes / 4 );

#if 0
	/* benchmark the read speed of the card's dma */
	if ( mgaglx.dmaDriver >= 2 ) {
		DmaBenchmark( (unsigned int)bufferPhysical, bufferBytes / 4 );
 		DmaBenchmark( (unsigned int)bufferPhysical, bufferBytes / 4 );
 		DmaBenchmark( (unsigned int)bufferPhysical, bufferBytes / 4 );
 		DmaBenchmark( (unsigned int)bufferPhysical, bufferBytes / 4 );
 		DmaBenchmark( (unsigned int)bufferPhysical, bufferBytes / 4 );
#if 0
	  /* switching to agp x2 does not work yet */
	  	if (use_agp) { 
		  OptimizeDMA3(); /* enable agp x2 */
 		  DmaBenchmark( (unsigned int)bufferPhysical, bufferBytes / 4 );
   		  DmaBenchmark( (unsigned int)bufferPhysical, bufferBytes / 4 );
		}
#endif	  
	        OptimizeDMA();
 		DmaBenchmark( (unsigned int)bufferPhysical, bufferBytes / 4 );
 		DmaBenchmark( (unsigned int)bufferPhysical, bufferBytes / 4 );
 		DmaBenchmark( (unsigned int)bufferPhysical, bufferBytes / 4 );
 		DmaBenchmark( (unsigned int)bufferPhysical, bufferBytes / 4 );
 		DmaBenchmark( (unsigned int)bufferPhysical, bufferBytes / 4 );
	}
#endif

	/* always leave enough room for a X server setup and DWGSYNC after overflow checks */
	mgaDmaBufferDwords = ( bufferBytes >> 3 ) - 30;

	/* setup the two buffers that will be ping-ponged */
	dmaBuffers[0] = malloc(sizeof(mgaDma_buffer));
	memset(dmaBuffers[0], '\0', sizeof(mgaDma_buffer));
	dmaBuffers[0]->buf_head = (mgaUI32 *)bufferVirtual;
	dmaBuffers[0]->physical = bufferPhysical;
	
	dmaBuffers[1] = malloc(sizeof(mgaDma_buffer));
	memset(dmaBuffers[1], '\0', sizeof(mgaDma_buffer));
	dmaBuffers[1]->buf_head = (mgaUI32 *)bufferVirtual + bufferBytes/8;
	dmaBuffers[1]->physical = bufferPhysical + bufferBytes/2;

	activeDmaBuffer = 0;
	dma_buffer = dmaBuffers[ activeDmaBuffer ];	
}


#ifdef USE_GART
static int AllocateGARTMemory(size_t size) 
{
  int i, j, pages = (size + 4095) / 4096; 
  
  gartfd = open("/dev/gart", O_RDWR);
  if (gartfd == -1)
    {	
      mgaMsg(1, "unable to open /dev/gart: %s\n", sys_errlist[errno]);
      return -1;
    }
  
  if (ioctl(gartfd, GARTIOCINFO, &gartinf) != 0)
    {
      mgaMsg(1, "error doing ioctl(GARTIOCINFO): %s\n", sys_errlist[errno]);
      close(gartfd);
      return -1;
    }
  
  gartbuf = mmap(NULL, gartinf.size * 0x100000, PROT_READ | PROT_WRITE, MAP_SHARED, gartfd, 0);
  if (gartbuf == (void *)0xffffffff)
    {
      mgaMsg(1, "mmap() on /dev/gart failed: %s\n", sys_errlist[errno]);
      close(gartfd);
      return -1;
    }
  
     
  for (i=0; i<pages; i++)
    {
      if (ioctl(gartfd, GARTIOCINSERT, &i))
	{	
	  /* we dont have enough memory, free the already allocated pages */
	  for (j=0;j <i; j++)
	    ioctl(gartfd, GARTIOCREMOVE, &j);
	  
	  mgaMsg(1, "GART: allocation of %i pages failed\n", pages);
	  
	  return -1;
	}	
    }
  
  return 0;
}
#endif

/*
 * AllocateSystemMemory
 * Looks at environment variables to determine if a block
 * of physical memory has been left for graphics after the
 * memory available to the kernel.
 * System memory can be used for dma command buffers or
 * textures.
 */
static void AllocateSystemMemory( void ) {
  int		fd;
  
  /* determine total size of buffer */
  sysmemBytes = atoi( getenvSafe("GLX_MGA_DMASIZE") );
  if ( !sysmemBytes ) {
    mgaMsg(1,"GLX_MGA_DMASIZE not set, skipping physical allocation\n" );
    return;
  }
  sysmemBytes *= 0x100000;

#ifdef USE_GART
  if (!AllocateGARTMemory(sysmemBytes))
    {
      sysmemPhysical = gartinf.physical;
      sysmemVirtual = (unsigned char *)gartbuf;
      sysmemHeap = mmInit( sysmemBytes );
      					 
      mgaMsg( 1, "AGP Aperture: %p\n", sysmemPhysical );
      mgaMsg( 1, "sysmemSize: %p\n", sysmemBytes );
      
      use_agp = PDEA_pagpxfer_enable /* | PDEA_primnostart_enable */ ;
      mgaMsg(1, "use_agp = %x\n", use_agp);
      switch (gartinf.agpmode)
      {
	case 1 :OptimizeDMA2(); /* turns on AGP x1 */
		break;
	case 2 :OptimizeDMA3(); /* turns on AGP x2 */
		break;
	default: mgaError("gart: agpmode %d ?!\n", gartinf.agpmode);
      }
      return;
    }	
#endif

  /* mgaglx.dmaDriverADR should be set to a value >= the mem= kernel parm */
  sysmemPhysical = atoi( getenvSafe("GLX_MGA_DMAADR") );
  if ( sysmemPhysical < 16 ) {
    mgaMsg( 1, "unlikely GLX_MGA_DMAADR=%i, skipping physical allocation\n", bufferPhysical );
    return;
  }
  sysmemPhysical *= 0x100000;


  
  /* FIXME!!!: should check sysmemPhysical against /proc/meminfo */
  
  fd = open( "/dev/mem", O_RDWR );
  if ( fd < 0 ) {
    mgaMsg( 1, "failed to open /dev/mem\n" );
    return;
  }
  
  sysmemVirtual = (unsigned char *) mmap( NULL, sysmemBytes, PROT_READ | PROT_WRITE,
					  MAP_SHARED, fd, (off_t)sysmemPhysical );
  if ( sysmemVirtual == MAP_FAILED ) {
    mgaMsg( 1, "failed to mmap sysmem\n" );
    close( fd );
    return;
  }

  /* FIXME: should verify the memory exists with read / write test */
  
  /* set to write combining */
  /* SetWriteCombining( sysmemPhysical, sysmemBytes ); */
  
  /* create a heap */
  sysmemHeap = mmInit( sysmemBytes );
  
  mgaMsg( 1, "sysmemPhysical: %p\n", sysmemPhysical );
  mgaMsg( 1, "sysmemVirtual: %p\n", sysmemVirtual );
  mgaMsg( 1, "sysmemSize: %p\n", sysmemBytes );
  
}

/*
 * ChooseTextureHeap
 * Determine if textures should be stored in the cardHeap or the
 * sysmemHeap.
 */
static void ChooseTextureHeap( void ) {
	/* share textures on the card memory until proven otherwise */
	textureHeap = cardHeap;
      	textureHeapVirtual = (unsigned char *)vgaLinearBase;
      	textureHeapPhysical = 0;

	/* if we don't have a system memory heap, textures MUST be on the card */
	if ( !sysmemHeap ) {
		mgaMsg( 1, "No sysmemHeap, textures must be stored on card\n" );
		return;
	}
	
	/* see if we should use the system heap for textures or just dma commands */
	if ( !atoi( getenvSafe("GLX_MGA_SYSTEMTEXTURE") ) ) {
		mgaMsg( 1, "GLX_MGA_SYSTEMTEXTURE not set, textures will be stored on card\n" );
		return;
	}

	/* make sure there is some memory left for textures */
	if ( sysmemBytes < bufferBytes + 1024*1024 ) {
		mgaMsg( 1, "sysmemBytes < bufferBytes + 1meg, textures will be stored on card\n" );
		return;
	}
	
	textureHeap = sysmemHeap;
      	textureHeapVirtual = sysmemVirtual;
      	textureHeapPhysical = sysmemPhysical;

	mgaMsg( 1, "Texturing from sysmemHeap\n" );
}


/*
===================
mgaDmaInit

===================
*/
void mgaDmaInit(void) {
  
  	mgaUI32 devctrl = pcibusRead( MGAPciTag, 0x04 );
    	mgaMsg( 1, "devctrl = %08x\n", devctrl);
 
        use_agp = 0;
    
	/* prepare to set write combining */
	MgaOpenMTRR();
	
	/* set write combining on the framebuffer */
	SetWriteCombining( xf86AccelInfoRec.ServerInfoRec->physBase,
		xf86AccelInfoRec.ServerInfoRec->physSize );

	/* get some system memory and make it write combining if we can */
	AllocateSystemMemory();
	mgaMsg(1,"use_agp: %i\n", use_agp );

	/* set up the matrox pdma memory window */	
	MapPseudoDmaWindow();
	
	/* read the command environment variable */
	mgaglx.dmaDriver = atoi( getenvSafe("GLX_MGA_DMA") );
	mgaMsg(1,"mgaDmaInit: GLX_MGA_DMA = %i\n", mgaglx.dmaDriver );

	/* setup the two command buffers in the apropriate memory space */
	AllocateCommandBuffers();
	
	/* check for using a PCI texture heap */
	ChooseTextureHeap();
}

/*===================================================*/

static  void mgaFlushPseudoDma( void ) {
	mgaUI32	save;
	mgaUI32	*code;
	mgaUI32	*p;
	int		count;
	
	code = dma_buffer->buf_head;
	count = dma_buffer->number_of_dwords;

	save = INREG(MGAREG_OPMODE);
	OUTREG( MGAREG_OPMODE,(save & OM_dmamod_MASK) | OM_dmamod_general );
  
	 p = pseudoDmaVirtual;
	while ( count-- ) {
		*p++ = *code++;
	}
  	
	OUTREG(MGAREG_OPMODE,save);
}

static void UnlockRegisters( void ) {
	mprotect( MGAMMIOBase, 0x3000, PROT_READ | PROT_WRITE );
	registersLocked = 0;
}

static void RegisterAccessSignalHandler( int signal ) {
	if ( !registersLocked ) {
		mgaMsg( 10, "RegisterAccessSignalHandler() without registersLocked\n" );
		mgaError( "RegisterAccessSignalHandler() without registersLocked\n" );
		exit( 1 );	/* FIXME: is there something better to do here? */
	}
	
	/* someone has tried to access hardware registers, so make sure dma is completed */
	mgaMsg( 10, "RegisterAccessSignalHandler()\n" );
	mgaglx.c_signals++;
	waitForDmaCompletion();
	mgaMsg( 10, "Leaving RASH()\n" );
}

static void LockRegisters( void ) {
	/* cause a SIGSEGV if the X server tries to write a hardware register */
	mprotect( MGAMMIOBase, 0x3000, /* PROT_NONE */ PROT_READ );
	signal( SIGSEGV, RegisterAccessSignalHandler );
	registersLocked = 1;
}

int	xchangeDummy;

#ifndef NO_MTRR
void FlushWriteCombining( void ) {
	__asm__ volatile( " push %%eax ; xchg %%eax, %0 ; pop %%eax" : : "m" (xchangeDummy));
	__asm__ volatile( " push %%eax ; push %%ebx ; push %%ecx ; push %%edx ; movl $0,%%eax ; cpuid ; pop %%edx ; pop %%ecx ; pop %%ebx ; pop %%eax" : /* no outputs */ :  /* no inputs */ );
}
#endif

static  void mgaFlushRealDma( void ) {
	int		count;

	mgaMsg( 10, "mgaFlushRealDma()\n" );

#ifndef NO_MTRR
	/* make sure any write combining data is flushed */
	FlushWriteCombining();
#endif	
	/* if we are using a card memory buffer, do a read to guarantee the store 
	buffer is flushed */
	xchangeDummy = dma_buffer->buf_head[dma_buffer->number_of_dwords];

	count = dma_buffer->number_of_dwords;
	dmaEnd = dma_buffer->physical + count*4;

	OUTREG( MGAREG_PRIMADDRESS, dma_buffer->physical );
	OUTREG( MGAREG_PRIMEND, dmaEnd | use_agp);

	/* if we are going to run async, throw a signal if X tries to use
	the hardware registers */
	if ( mgaglx.dmaDriver == 3 ) {
		LockRegisters();
	}
}

/*
 * mgaDmaFlush
 * Send all pending commands off to the hardware.
 * If we are running async, the hardware will be drawing
 * while we return to do other things.
 */
void mgaDmaFlush(void) {
	int		start, end;
	DMALOCALS;

	/* if the buffer is empty, don't do anything */
	if ( dma_buffer->number_of_dwords == 0 ) {
		return;
	}

	mgaMsg( 10, "mgaDmaFlush()\n" );
	mgaglx.c_dmaFlsuh++;

	/* wait for the last buffer to complete */
	if ( waitForDmaCompletion() ) {
		mgaglx.fullUtilization = 1;
	} else {
		mgaglx.fullUtilization = 0;
	}

	/* set the DWGSYNC register to our magic value */
	OUTREG( MGAREG_DWGSYNC, SYNC_DMA_BUSY );
	while ( INREG( MGAREG_DWGSYNC ) != SYNC_DMA_BUSY ) {
		/* X might have left something drawing */
	}

	if ( dma_buffer->number_of_dwords > largestDmaBuffer ) {
		largestDmaBuffer = dma_buffer->number_of_dwords;
	}

	/* add the commands at the end of the buffer to go back to drawing on the
	front buffer the way the X server expects */
	memcpy(dma_buffer->buf_head + dma_buffer->number_of_dwords, 
	mgaFrontBuffer->Setup, sizeof(mgaUI32) * mgaFrontBuffer->SetupSize);
	dma_buffer->number_of_dwords += mgaFrontBuffer->SetupSize;
	
	/* add a draw sync command at the end so we can tell when dma is done */
	MGADMAGETPTR( 4 );
	DMAOUTREG( MGAREG_DMAPAD, 0 );
	DMAOUTREG( MGAREG_DMAPAD, 0 );
	DMAOUTREG( MGAREG_DMAPAD, 0 );
	DMAOUTREG( MGAREG_DWGSYNC, 0 );
	DMAADVANCE();

	/* collect timing information if we are going syncronously */
	if ( mgaglx.dmaDriver != 3 ) {
 		start = usec();
 	} else {
 		start = end = 0;
 	}
 	
 	if ( mgaglx.dmaDriver < 2 ) {
	 	mgaFlushPseudoDma();
	 } else {
 		mgaFlushRealDma();
 	}
 
	if ( mgaglx.dmaDriver == 2 ) {
		/* wait until the dma completes */
		waitForDmaCompletion();
	}

 	if ( mgaglx.dmaDriver != 3 ) {
		end = usec();
	}

	mgaMsg(10, "flushmode %i, buffer %i: %i dwords %i usec\n", 
		mgaglx.dmaDriver,  activeDmaBuffer, dma_buffer->number_of_dwords,
		end - start );	

	/* swap to using the other buffer */
	activeDmaBuffer ^= 1;
	dma_buffer = dmaBuffers[ activeDmaBuffer ];

	dma_buffer->number_of_dwords = 0;
	dma_buffer->registerOffset = 4;
	
	/* make sure the current state is set at the beginning of the next buffer */
	mgaglx.needEnter3D = 1;
}

/*
 * mgaDmaOverflow
 * This is called when MGADMAGETPTR is at the end of the buffer
 */
void mgaDmaOverflow( int newDwords ) {
	mgaMsg( 10, "mgaDmaOverflow(%i)\n", newDwords );

	/* flush all the current commands so we will have another empty buffer */
	mgaDmaFlush();

	/* reset the current drawing state at the beginning of the buffer
	we can't rely on the normal primitive process to do this, because an
	overflow can occur in the middle of a rendering primitive */
	ENTER3D;

	mgaglx.c_overflows++;
	if ( newDwords > mgaDmaBufferDwords ) {
		mgaMsg( 1, "HORRIBLE ERROR: %i > mgaDmaBufferDwords\n", newDwords);
	}
}

/*
 * mgaDmaFinish
 */
void mgaDmaFinish( void ) {
	mgaDmaFlush();
	waitForDmaCompletion();
}

/*
 * mgaWaitDrawingEngine
 * This will not return until the drawing engine has completed
 * drawing pixels and it is safe to read or write the framebuffer
 * for software rendering.
 */
int mgaWaitDrawingEngine( void ) {
	/* note this for the performance block display */
	mgaglx.c_drawWaits++;

	/* make sure all pending dma has completed */
	mgaDmaFinish();

	return 0;
}


/*
 * mgaDmaExecute
 * Add a block of data to the dma buffer
 */
void mgaDmaExecute( mgaUI32 *code, int dwords ) {
	if( (mgaDmaBufferDwords - dma_buffer->number_of_dwords) < dwords) {
		 mgaDmaFlush();
	}
	memcpy( dma_buffer->buf_head + dma_buffer->number_of_dwords, 
		code, sizeof(mgaUI32) * dwords );
	dma_buffer->number_of_dwords += dwords;
}


/*
 * mgaSet3DState
 * Add dma commands to prepare for drawing in the current context.
 * This is done whenever a primitive is going to be drawn and the
 * current state is invalid due to a change in functionality,
 * a context switch, or a dma buffer switch.
 * Returns 0 if 3D drawing should not be performed for some reason.
 */
int mgaSet3DState( void ) {
	mgaglx.needEnter3D = 0;
	if ( !mgaDB ) {
		mgaMsg( 1, "mgaSet3DState: !mgaDB\n" );
		return 0;
	}
	if ( !mgaCtx ) {
		mgaMsg( 1, "mgaSet3DState: !mgaCtx\n" );
		return 0;
	}

	/* set the back buffer and depth buffer */
	mgaDmaExecute( mgaDB->Setup,mgaDB->SetupSize );
	
	return 1;
}
