/*
  GLX Hardware Device Driver for Matrox Millenium G200
  Copyright (C) 1999  Stephen Crowley (crow@debian.org)
  
  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
  
 original by Jeff Hartmann <slicer@ionet.net>
  6/16/99: rewrite by John Carmack <johnc@idsoftware.com>
*/

#ifndef MGADMA_H
#define MGADMA_H

#include "mgacommon.h"

/* the g200 general purpose dma stream operates in blocks of five
dwords that program four registers at a time.  One index dword
specifies the registers to program with the following four dwords of data.

ADR0 | (ADR1<<8 ) | (ADR2<<16) | (ADR3<<24)
DATA0
DATA1
DATA2
DATA3
....
*/

#define DWGREG0 0x1c00
#define DWGREG0_END 0x1dff
#define DWGREG1 0x2c00
#define DWGREg1_END 0x2dff

/* page 407 in the specs, 4-29 */
#define ISREG0(r) (r >= DWGREG0 && r<= DWGREG0_END)
#define ADRINDEX0(r) (mgaUI8)((r - DWGREG0) >> 2)
#define ADRINDEX1(r) (mgaUI8)(((r - DWGREG1) >> 2) | 0x80)
#define ADRINDEX(r) (ISREG0(r) ? ADRINDEX0(r) : ADRINDEX1(r)) 

#define MGA_ADDRGEN(i0, i1, i2, i3) \
 (ADRINDEX(i0) | \
 (ADRINDEX(i1) << 8) | \
 (ADRINDEX(i2) << 16) | \
 (ADRINDEX(i3) << 24))

/* the init function will try to get real physical memory for dma, but
it will fall back to just mallocing a block of virtual memory and using
pseudo dma if nothing else is available */
void mgaDmaInit(void);

/* Add dma commands to prepare for drawing in the current context.
This is done whenever a primitive is going to be drawn and the
current state is invalid due to a change in functionality,
a context switch, or a dma buffer switch.
Returns 0 if 3D drawing should not be performed for some reason. */
int mgaSet3DState( void );

/* this macro should be used before any code that will emit drawing commands */
#define ENTER3D \
	if ( mgaglx.needEnter3D ) { \
		if ( !mgaSet3DState() ) { 		\
			return; /* can't do 3D drawing for some reason */ \
		}		\
	}


/* a flush command will guarantee that all data added to the dma buffer
is on its way to the card, and will eventually complete with no more
intervention.  If running with pseudo dma, this will be the same as a finish
call, but if async dma is active then the card will be executing the commands
while the cpu is doing other work.  A protected memory region keeps X server
interaction with the hardware registers safe. */
void mgaDmaFlush(void);

/* the overflow function is called when a block can't be allocated
in the current dma buffer.  It flushes the current buffer and
records some information */
void mgaDmaOverflow(int newDwords);

/* a finish command will guarantee that all dma commands have actually
been consumed by the card.  Note that there may still be a couple primitives
that have not yet been executed out of the internal FIFO, so this does not
guarantee that the drawing engine is idle. */
void mgaDmaFinish( void );

/* a mgaWaitDrawingEngine command will guarantee that the framebuffer
is safe to read or write for software rendering */
int mgaWaitDrawingEngine( void );

/* pre-generated blocks of register commands can be added to the dma buffer
with mgaDmaExecute().  This is used to set the state registers and texture
registers.  The dwords size MUST be a multiple of 5, pad the index bytes with
nops (0x15) and extra data dwords if needed */
void mgaDmaExecute(mgaUI32 *code, int dwords);

/* for routines that need to add a variable stream of register writes, use
the MGADMAGETPTR() / DMAOUTREG() / DMAADVANCE() interface.  Call
mgaDmaGetPtr() with a length in dwords that is guaranteed to be greater
than what you will be writing.  Using a large value, like 100, does not
waste anything.  Program all the registers you need with DMAOUTREG(),
or the MGADMA_* macros, which use DMAOUTREG().  When you have finished,
call DMAADVANCE(), which will add any necessary padding and commit the
data to the dma buffer 

DMALOCALS must be included at the top of all functions that use these
macros to declare temporary variables.

*/

typedef struct _dma_buffer {
   mgaUI32 	*buf_head;
   mgaUI32	physical;
   mgaUI32	number_of_dwords;
   mgaUI32	indexBase;
   mgaUI32	registerOffset;
} mgaDma_buffer;

/* cardHeap is the 8 / 16 / 32 megs of memory on the video card */
extern	memHeap_t	*cardHeap;
extern	mgaUI32		cardPhysical;
extern	unsigned char	*cardVirtual;

/* sysmemHeap is system memory we have been given after the area used by the kernel */
extern	memHeap_t	*sysmemHeap;
extern	mgaUI32		sysmemPhysical;			/* 0 if we don't have a physical mapping */
extern	unsigned char	*sysmemVirtual;

/* textureHeap will point to either cardHeap or sysmemHeap */
extern	memHeap_t	*textureHeap;
extern	mgaUI32		textureHeapPhysical;		/* 0 if we aren't using PCI texturing */
extern	unsigned char	*textureHeapVirtual;		/* correct for either local or PCI heaps */


extern	mgaDma_buffer	*dma_buffer;
extern	mgaUI32	mgaDmaBufferDwords;
extern	mgaUI32	ValueOffset[100];
extern	mgaUI32	IndexOffset[100];


#define DMALOCALS int outcount; mgaUI32 *dma_ptr; unsigned char tempIndex[4]

#define MGADMAGETPTR(length) \
   if( (mgaDmaBufferDwords - dma_buffer->number_of_dwords) < length) \
     mgaDmaOverflow(length); \
   dma_ptr = (dma_buffer->buf_head + dma_buffer->number_of_dwords); \
   outcount = 0;

#define DMAOUTREG(reg ,val) do {       			\
tempIndex[outcount]=ADRINDEX(reg); \
dma_ptr[1+outcount]=val;		\
if ( ++outcount == 4 ) { \
outcount = 0; \
dma_ptr[0]=*(int *)tempIndex; \
dma_ptr+=5;	\
dma_buffer->number_of_dwords += 5;	       		\
} \
}while(0)

#define DMAADVANCE() do {					\
if ( outcount & 3 ) { \
while(outcount & 3) {	\
	tempIndex[outcount++]=0x15; \
}							\
dma_ptr[0]=*(int *)tempIndex; \
dma_buffer->number_of_dwords += 5;	       		\
} \
}while(0)


#endif
