/*
 * GLX Hardware Device Driver for Matrox Millenium G200
 * Copyright (C) 1999 Wittawat Yamwong
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included
 * in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
 * WITTAWAT YAMWONG, OR ANY OTHER CONTRIBUTORS BE LIABLE FOR ANY CLAIM, 
 * DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR 
 * OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE 
 * OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 *
 *
 *    Wittawat Yamwong <Wittawat.Yamwong@stud.uni-hannover.de>
 */

/* $Id: mgatex.c,v 1.1 1999/07/31 07:07:55 crow Exp $ */

#include "xsmesaP.h"
#include <stdlib.h>

#include <GL/gl.h>

#include "g200_mac.h"
#include "mm.h"
#include "mgalib.h"
#include "mgatex.h"
#include "mgalog.h"

extern int use_agp;

static void ReplicateMesaTexState(mgaTextureObjectPtr t,
                                  struct gl_texture_object *mesatex)
{
  mgaSetTexWrapping(t,mesatex->WrapS,mesatex->WrapT);
  mgaSetTexFilter(t,mesatex->MinFilter,mesatex->MagFilter);
  mgaSetTexBorderColor(t,mesatex->BorderColor);
}

mgaTextureObjectPtr mgaCreateTexObj(mgaContextPtr ctx,
				    struct gl_texture_object *tObj)
{
  mgaTextureObjectPtr t;
  mgaUI32 *p;

  t = (mgaTextureObjectPtr) calloc(1,sizeof(struct mga_texture_object_t));
  if (!t)
    return NULL;
  t->magic = MGA_TEXTURE_OBJECT_MAGIC;
  t->globj = tObj;
  p = t->TexReg;
  p[0] = MGA_TEXREG_0;
  p[5] = MGA_TEXREG_5;
  p[10] = MGA_TEXREG_10;
  p[15] = MGA_TEXREG_15;
  p[MGA_TEXREG_TEXBORDERCOL] = MGAPACKCOLOR8888(255,0,0,0);
  p[MGA_TEXREG_TEXCTL] = TMC_takey_1 | TMC_tamask_0;
  p[MGA_TEXREG_TEXCTL2] = TMC_ckstransdis_enable;
  p[MGA_TEXREG_TEXFILTER] = TF_minfilter_nrst | TF_magfilter_nrst |
    TF_filteralpha_enable | (0x10 << TF_fthres_SHIFT);
  p[MGA_TEXREG_TEXHEIGHT] = 0;
  p[MGA_TEXREG_TEXWIDTH] = 0;
  p[MGA_TEXREG_TEXTRANS] = 0xffff;
  p[MGA_TEXREG_TEXTRANSHIGH] = 0xffff;
  p[MGA_TEXREG_TEXORG] = 0;
  p[MGA_TEXREG_TEXORG1] = 0;
  p[MGA_TEXREG_TEXORG2] = 0;
  p[MGA_TEXREG_TEXORG3] = 0;
  p[MGA_TEXREG_TEXORG4] = 0;

  ReplicateMesaTexState(t,tObj);
  if (ctx->gl_ctx->Light.Model.ColorControl == GL_SEPARATE_SPECULAR_COLOR)
    p[MGA_TEXREG_TEXCTL2] |= TMC_specen_enable;

  t->age  = ++ctx->current_texture_age;
  t->next = ctx->TexObjList;
  ctx->TexObjList = t;
  return t;
}

int mgaDestroyTexObj(mgaContextPtr ctx, mgaTextureObjectPtr t)
{
  int i;
  mgaTextureObjectPtr p,prev;

  if (!t)
    return 0;
  if (t->magic != MGA_TEXTURE_OBJECT_MAGIC)
    return -1;
  for (i=0; i < MGA_TEX_MAXLEVELS; i++) {
    mgaDestroyBuffer(t->Tex[i]);
  } 
  mgaMsg(10,"mgaDestroyTexObj()\n");
  /* mmDumpMemInfo( textureHeap ); */
  /* remove from the list */
  p = ctx->TexObjList;
  prev = NULL;
  while (p) {
    if (p == t) {
      if (prev)
	prev->next = t->next;
      else
	ctx->TexObjList = t->next;
      break;
    }
    prev = p;
    p = p->next;
  }

  t->magic = 0;
  free(t);
  return 0;
}

void mgaSwapOutTexObj(mgaTextureObjectPtr t)
{
  mgaBufferPtr buf;
  void *v;
  int i;
  
  if (!t) return;
  if (t->swapped_out) return;
  
  mgaMsg(10,"  Swapping out texture.\n");
  for (i=0; i < MGA_TEX_MAXLEVELS; i++) {
    buf = t->Tex[i];
    if (VALID_MGA_BUFFER(buf)) {
      if (buf->CachedContent) {
      	/* we already have a copy in virtualMemory */
	  v = buf->CachedContent;
      } else {
      	/* copy the texture from the textureHeap memory to virtual memory */
	  v = (void *) malloc(buf->Size);
	  mgaMsg(10," out Src=%p Dst=%p\n", buf->BufAddr, v);
	  memcpy(v, buf->BufAddr, buf->Size);
	  buf->CachedContent=v;
      }
      mmFreeMem(buf->MemBlock);
      buf->BufAddr = v;
      buf->MemBlock= NULL;
      buf->UseCardMemory = 0;
    }
    else
      t->Tex[i] = NULL;
  } 
  t->swapped_out = 1;
}

int mgaSwapOutOldestTexObj()
{
  mgaTextureObjectPtr t, oldest;
  unsigned long old;
   
  old = 0x7fffffff;
  oldest = NULL;
  for (t = mgaCtx->TexObjList; t ; t=t->next)
    if (!t->swapped_out && (t->age < old)) {
      old = t->age;
      oldest = t;
    }
  if (oldest)
    mgaSwapOutTexObj(oldest);
  else
  {
    /* This is a severe error that should not happen! */
    mgaError("  No Texture to swap out -> Out of Memory!\n"); 
    for (t = mgaCtx->TexObjList; t ; t=t->next) {
      mgaMsg(1,"  texture  swapped:%d age:%8d\n", t->swapped_out, t->age);
    }
    mmDumpMemInfo( textureHeap ); 
    return -1;
  }
  return 0;
}


int mgaSwapInTexObj(mgaTextureObjectPtr t)
{
  mgaBufferPtr buf;
  int i, ofs;
  void *v;
  
  mgaMsg(10,"  Swapping in texture.\n");
  mgaglx.c_textureSwaps++;
  mgaWaitDrawingEngine();	/* make sure no drawing is going on */
  for (i=0; i < MGA_TEX_MAXLEVELS; i++) {
    buf = t->Tex[i];
    if (buf) {
      if (!VALID_MGA_BUFFER(buf)) {
        mgaError("  swapped out buffer level %d not valid.\n",i);
        t->Tex[i] = NULL;
        return 0;
      }
      if (!buf->UseCardMemory)
      { 
      /* allocate textureHeap memory, swapping out stuff until we can */
        while ((buf->MemBlock = mmAllocMem( textureHeap, buf->Size,5, 0 )) == 0)
          if (mgaSwapOutOldestTexObj()) return -1;
	
        v = buf->BufAddr;	/* virtual memory */
        ofs = buf->MemBlock->ofs;
        buf->Setup[MGA_SETUP_DSTORG] = ofs;
        buf->BufAddr = textureHeapVirtual + ofs;	/* texture memory */
        mgaMsg(10," in  Src=%p Dst=%p\n", v, buf->BufAddr);
        memcpy(buf->BufAddr, v, buf->Size);
        buf->UseCardMemory = 1;
      }
      switch (i) {
        case 0: t->TexReg[MGA_TEXREG_TEXORG] = textureHeapPhysical + buf->Setup[MGA_SETUP_DSTORG] ; 
        	/* turn on PCI reads if we are using PCI texturing */
        	if ( textureHeapPhysical ) {
        		t->TexReg[MGA_TEXREG_TEXORG] |= 1 | use_agp;
        	}
        	break;
        case 1: t->TexReg[MGA_TEXREG_TEXORG1]= textureHeapPhysical + buf->Setup[MGA_SETUP_DSTORG]; break;
        case 2: t->TexReg[MGA_TEXREG_TEXORG2]= textureHeapPhysical + buf->Setup[MGA_SETUP_DSTORG]; break;
        case 3: t->TexReg[MGA_TEXREG_TEXORG3]= textureHeapPhysical + buf->Setup[MGA_SETUP_DSTORG]; break;
        case 4: t->TexReg[MGA_TEXREG_TEXORG4]= textureHeapPhysical + buf->Setup[MGA_SETUP_DSTORG]; break;
      }
    }
  }                    
  t->swapped_out = 0;
  return 0;
}

void mgaSetTexWrapping(mgaTextureObjectPtr tex, GLenum s, GLenum t)
{
  if (s == GL_REPEAT) {
    tex->TexReg[MGA_TEXREG_TEXCTL] &= ~TMC_clampu_enable;
  } else {
    tex->TexReg[MGA_TEXREG_TEXCTL] |= TMC_clampu_enable;
  }
  if (t == GL_REPEAT) {
    tex->TexReg[MGA_TEXREG_TEXCTL] &= ~TMC_clampv_enable;
  } else {
    tex->TexReg[MGA_TEXREG_TEXCTL] |= TMC_clampv_enable;
  }
}

void mgaSetTexFilter(mgaTextureObjectPtr t, GLenum minf, GLenum magf)
{
  switch (minf) {
  case GL_NEAREST:
    MGA_SET_FIELD(t->TexReg[MGA_TEXREG_TEXFILTER],
		  TF_minfilter_MASK & TF_mapnb_MASK,
		  TF_minfilter_nrst);
    break;
  case GL_LINEAR:
    MGA_SET_FIELD(t->TexReg[MGA_TEXREG_TEXFILTER],
		  TF_minfilter_MASK & TF_mapnb_MASK,
		  TF_minfilter_bilin);
    break;
  case GL_NEAREST_MIPMAP_NEAREST:
    MGA_SET_FIELD(t->TexReg[MGA_TEXREG_TEXFILTER],
		  TF_minfilter_MASK & TF_mapnb_MASK,
		  TF_minfilter_mm1s);
    break;
  case GL_LINEAR_MIPMAP_NEAREST:
    MGA_SET_FIELD(t->TexReg[MGA_TEXREG_TEXFILTER],
		  TF_minfilter_MASK & TF_mapnb_MASK,
		  TF_minfilter_mm4s);
    break;
  case GL_NEAREST_MIPMAP_LINEAR:
    MGA_SET_FIELD(t->TexReg[MGA_TEXREG_TEXFILTER],
		  TF_minfilter_MASK & TF_mapnb_MASK,
		  TF_minfilter_mm2s);
    break;
  case GL_LINEAR_MIPMAP_LINEAR:
    MGA_SET_FIELD(t->TexReg[MGA_TEXREG_TEXFILTER],
		  TF_minfilter_MASK & TF_mapnb_MASK,
		  TF_minfilter_mm8s);
    break;
  default:
    mgaError("mgaSetTexFilter(): not supported min. filter %d\n",(int)minf);
    break;
  }

  switch (magf) {
  case GL_NEAREST:
    MGA_SET_FIELD(t->TexReg[MGA_TEXREG_TEXFILTER],
		  TF_magfilter_MASK,TF_magfilter_nrst);
    break;
  case GL_LINEAR:
    MGA_SET_FIELD(t->TexReg[MGA_TEXREG_TEXFILTER],
		  TF_magfilter_MASK,TF_magfilter_bilin);
    break;
  default:
    mgaError("mgaSetTexFilter(): not supported mag. filter %d\n",(int)magf);
    break;
  }
  
  /* See OpenGL 1.2 specification */
  if (magf == GL_LINEAR && (minf == GL_NEAREST_MIPMAP_NEAREST || 
                            minf == GL_NEAREST_MIPMAP_LINEAR))
  {
    /* c = 0.5 */
    MGA_SET_FIELD(t->TexReg[MGA_TEXREG_TEXFILTER],TF_fthres_MASK,
		  0x20 << TF_fthres_SHIFT);
  } else {
    /* c = 0 */
    MGA_SET_FIELD(t->TexReg[MGA_TEXREG_TEXFILTER],TF_fthres_MASK,
		  0x10 << TF_fthres_SHIFT);
  }
}

void mgaSetTexBorderColor(mgaTextureObjectPtr t, GLint color[4])
{
  t->TexReg[MGA_TEXREG_TEXBORDERCOL] = 
    MGAPACKCOLOR8888(color[0],color[1],color[2],color[3]);
}

void mgaUpdateTextureState(GLcontext *ctx, mgaTextureObjectPtr t)
{
  int LastLevel,i;
  int Update = 0;

  if (!t) {
    mgaError("mgaUpdateTextureState: software fallback, t == NULL\n");
    mgaCtx->SoftwareFallback |= FALLBACK_TEXTURE;
    return;
  }

  if (ctx->NewState & NEW_LIGHTING) {
    mgaUI32 old = t->TexReg[MGA_TEXREG_TEXCTL2];
    if (ctx->Light.Model.ColorControl == GL_SEPARATE_SPECULAR_COLOR) {
      t->TexReg[MGA_TEXREG_TEXCTL2] |= TMC_specen_enable;
    } else {
      t->TexReg[MGA_TEXREG_TEXCTL2] &= ~TMC_specen_enable;
    }
    if (old != t->TexReg[MGA_TEXREG_TEXCTL2])
      Update = 1;
  }

  if (ctx->NewState & NEW_DRVSTATE0) {
    if (!t->globj->Complete) {
      t->Complete = 0;
      t->NullTexture = 1;
    } else if (t->swapped_out) {
      t->Complete = 0;
    } else {
      t->Complete = 1; /* be optimistic */
      t->NullTexture = 0;
      switch (mgaCtx->TextureMode) {
      case GL_REPLACE:
	t->TexReg[MGA_TEXREG_TEXCTL] &= ~TMC_tmodulate_enable;
	t->TexReg[MGA_TEXREG_TEXCTL2] &= ~TMC_decalblend_enable;
	break;
      case GL_MODULATE:
	t->TexReg[MGA_TEXREG_TEXCTL] |= TMC_tmodulate_enable;
	t->TexReg[MGA_TEXREG_TEXCTL2] &= ~TMC_decalblend_enable;
	break;
      case GL_DECAL:
	t->TexReg[MGA_TEXREG_TEXCTL] &= ~TMC_tmodulate_enable;
	t->TexReg[MGA_TEXREG_TEXCTL2] |= TMC_decalblend_enable;
	break;
      default:
	mgaError("mgaUpdateTexObject(): not supported texture mode %d\n",
	       mgaCtx->TextureMode);
	t->Complete = 0;
      }

      if (t->globj->MinFilter != GL_NEAREST && 
	  t->globj->MinFilter != GL_LINEAR) {
	LastLevel = t->globj->P;
	if (LastLevel >= MGA_TEX_MAXLEVELS)
	  LastLevel = MGA_TEX_MAXLEVELS-1;
      } else {
	LastLevel = 0;
      }
      for (i=0; i <= LastLevel; i++) {
	if (!VALID_MGA_BUFFER(t->Tex[i])) {
	  mgaMsg(1,"mgaUpdateTexObj(): level %d/%d not valid.\n",i,LastLevel);
	  t->Complete = 0;
	  break;
	}
      }
      MGA_SET_FIELD(t->TexReg[MGA_TEXREG_TEXFILTER],TF_mapnb_MASK,
		    LastLevel << TF_mapnb_SHIFT);
      Update = 1;
    }
  }

  if (Update && t->Complete) {

    if (t->UsePalette == MGA_USE_PALETTE) {
      
      if (ctx->Texture.SharedPalette) {
	if (mgaCtx->GlobalPaletteUpdated)
          mgaLoadTexturePalette(mgaCtx->GlobalPalette,0,256);
	mgaCtx->GlobalPaletteUpdated = 0;
      }
      else {
        mgaLoadTexturePalette(t->Palette,0,256);
	mgaCtx->GlobalPaletteUpdated = 1;
      }
    }		      

    mgaDmaExecute(t->TexReg,MGA_TEXREG_SIZE);
  }
  /* DO NOT put checking code here. Please put it before the last if. */

  if ((!ctx->Texture.Enabled) || t->Complete || t->NullTexture) {
    mgaCtx->SoftwareFallback &= ~FALLBACK_TEXTURE;
  } else {
    mgaMsg(1,"mgaUpdateTextureState: software fallback, Texture.Enabled && !t->Complete\n");
    mgaCtx->SoftwareFallback |= FALLBACK_TEXTURE;
  }
}

static int CopyImage(mgaTextureObjectPtr t,int level, int intFormat,
		     const struct gl_texture_image *image)
{
  mgaBufferPtr buf;
  mgaUI32 *p;
  int CacheFlag;

  mgaMsg(10,"CopyImage():\n");
  mgaglx.c_textureSwaps++;
   mgaWaitDrawingEngine();	/* make sure no drawing is going on */
 
  /* if the texture object already is swapped out we
   * don't need to request a CardMemory buffer! */
  CacheFlag = (t->swapped_out?MGA_TEXCACHE:0);

  if (level < MGA_TEX_MAXLEVELS) {
    mgaDestroyBuffer(t->Tex[level]);
    t->Tex[level] = NULL;
    
    if (image->Border != 0) {
      mgaError("Not supported texture border %d.\n",image->Border);
      return -1;
    }
    switch (image->Format) {
    case GL_RGB:
      buf = mgaCreateBuffer(MGA_TEXBUFFER | MGA_TW16 | CacheFlag, 
                            0,image->Width2,image->Height2,image->Width2);
      mgaMsg(10,"  CopyRGB: %p %p %p\n", buf->MemBlock, buf->BufAddr, buf->CachedContent);      
      t->Tex[level] = buf;
      if (!buf) 
	return -1;    

      buf->refcount++;
      {
        mgaUI16 *dst = (mgaUI16 *)buf->BufAddr;
        mgaUI8  *src = (mgaUI8 *)image->Data;
        int i;
        for (i=buf->Width*buf->Height; i > 0; i--) {
          *dst = MGAPACKCOLOR565(src[0],src[1],src[2]);
          dst++;
          src += 3;
        }
      }
      break;
      
    case GL_RGBA:
      buf = mgaCreateBuffer(MGA_TEXBUFFER | MGA_TW12 | CacheFlag,
      			    0,image->Width2,image->Height2,image->Width2);
      mgaMsg(10,"  CopyRGBA: %p %p %p\n", buf->MemBlock, buf->BufAddr, buf->CachedContent);      
      t->Tex[level] = buf;
      if (!buf)
	return -1;
 
      buf->refcount++;
      {
        mgaUI16 *dst = (mgaUI16 *)buf->BufAddr;
        mgaUI8  *src = (mgaUI8 *)image->Data;
        int i;
        for (i=buf->Width*buf->Height; i > 0; i--) {
          *dst = MGAPACKCOLOR4444(src[0],src[1],src[2],src[3]);
          dst++;
          src += 4;
        }
      }
      break;
      
    case GL_LUMINANCE:
      buf = mgaCreateBuffer(MGA_TEXBUFFER | MGA_TW16 | CacheFlag,
                            0,image->Width2,image->Height2,image->Width2);
      mgaMsg(10,"  CopyL: %p %p %p\n", buf->MemBlock, buf->BufAddr, buf->CachedContent);      
      t->Tex[level] = buf;
      if (!buf)
	return -1;
      
      buf->refcount++;
      {
        mgaUI16 *dst = (mgaUI16 *)buf->BufAddr;
        mgaUI8  *src = (mgaUI8 *)image->Data;
        int i;
        for (i=buf->Width*buf->Height; i > 0; i--) {
          *dst = MGAPACKCOLOR565(src[0],src[0],src[0]);
          dst++;
          src++;
        }
      }
      break;

    case GL_INTENSITY:
      buf = mgaCreateBuffer(MGA_TEXBUFFER | MGA_TW12 | CacheFlag,
                            0,image->Width2,image->Height2,image->Width2);
      mgaMsg(10,"  CopyI: %p %p %p\n", buf->MemBlock, buf->BufAddr, buf->CachedContent);      
      t->Tex[level] = buf;
      if (!buf)
	return -1;
	
      buf->refcount++;
      {
        mgaUI16 *dst = (mgaUI16 *)buf->BufAddr;
        mgaUI8  *src = (mgaUI8 *)image->Data;
        int i;
        for (i=buf->Width*buf->Height; i > 0; i--) {
          *dst = MGAPACKCOLOR4444(src[0],src[0],src[0],src[0]);
          dst++;
          src++;
        }
      }
      break;
      
    case GL_LUMINANCE_ALPHA:
      buf = mgaCreateBuffer(MGA_TEXBUFFER | MGA_TW12 | CacheFlag,
                            0,image->Width2,image->Height2,image->Width2);
      mgaMsg(10,"  CopyLA: %p %p %p\n", buf->MemBlock, buf->BufAddr, buf->CachedContent);      
      t->Tex[level] = buf;
      if (!buf)
	return -1;
      
      buf->refcount++;
      {
        mgaUI16 *dst = (mgaUI16 *)buf->BufAddr;
        mgaUI8  *src = (mgaUI8 *)image->Data;
        int i;
        for (i=buf->Width*buf->Height; i > 0; i--) {
          *dst = MGAPACKCOLOR4444(src[0],src[0],src[0],src[1]);
          dst++;
          src+=2;
        }
      }
      break;

    case GL_COLOR_INDEX:
         buf = mgaCreateBuffer(MGA_TEXBUFFER | MGA_TW8 | CacheFlag,
                            0,image->Width2,image->Height2,image->Width2);
      mgaMsg(10,"  CopyIndex: %p %p %p\n", buf->MemBlock, buf->BufAddr, buf->CachedContent);      
      t->Tex[level] = buf;
      t->UsePalette = MGA_USE_PALETTE;
      if (!buf)
	return -1;
	
      buf->refcount++;
      memcpy(buf->BufAddr, image->Data, buf->Width*buf->Height);
      break;
      
    default:
      mgaError("Not supported texture format %d\n",image->Format);
      return -1;
    }
    
    /* buf did not fit in CardMemory so we swap out the whole object: */
    if (!buf->UseCardMemory && !t->swapped_out) {
      mgaMsg(10,"  swap out partially.\n");
      mgaSwapOutTexObj(t); 
    }
  }

  /* update texture registers */
  p = t->TexReg;
  switch (level) {
  case 0:
    MGA_SET_FIELD(p[MGA_TEXREG_TEXCTL],
		  TMC_tformat_MASK & TMC_tpitchlin_MASK & TMC_tpitch_MASK,
		  ((t->Tex[0]->Attrib & MGA_PF_MASK) >> 4)); /* tformat */
    if (image->WidthLog2 >= 3) {
      p[MGA_TEXREG_TEXCTL] |= (image->WidthLog2 - 3) << TMC_tpitch_SHIFT;
    } else {
      p[MGA_TEXREG_TEXCTL] |= TMC_tpitchlin_enable | 
	(image->Width2 << TMC_tpitchext_SHIFT);
    }
    p[MGA_TEXREG_TEXHEIGHT] =
      MGA_FIELD(TH_th,     image->HeightLog2 + 4-MGA_TEX_TBIT+MGA_TEX_QBIT)|
      MGA_FIELD(TH_rfh,    8 - image->HeightLog2 - MGA_TEX_QBIT+16) |
      MGA_FIELD(TH_thmask, image->Height2 - 1);
    p[MGA_TEXREG_TEXWIDTH] = 
      MGA_FIELD(TW_tw,     image->WidthLog2 + 4-MGA_TEX_SBIT+MGA_TEX_QBIT) |
      MGA_FIELD(TW_rfw,    8 - image->WidthLog2 - MGA_TEX_QBIT+16) |
      MGA_FIELD(TW_twmask, image->Width2 - 1);
    p[MGA_TEXREG_TEXORG] = textureHeapPhysical + t->Tex[0]->Setup[MGA_SETUP_DSTORG];
    	/* turn on PCI reads if we are using PCI texturing */
    	if ( textureHeapPhysical ) {
    		p[MGA_TEXREG_TEXORG] |= 1 | use_agp;
    	}
    break;
    
  case 1:
    p[MGA_TEXREG_TEXORG1] = textureHeapPhysical + t->Tex[1]->Setup[MGA_SETUP_DSTORG];
    break;

  case 2:
    p[MGA_TEXREG_TEXORG2] = textureHeapPhysical + t->Tex[2]->Setup[MGA_SETUP_DSTORG];
    break;

  case 3:
    p[MGA_TEXREG_TEXORG3] = textureHeapPhysical + t->Tex[3]->Setup[MGA_SETUP_DSTORG];
    break;

  case 4:
    p[MGA_TEXREG_TEXORG4] = textureHeapPhysical + t->Tex[4]->Setup[MGA_SETUP_DSTORG];
    break;
  }
  
  return 0;
}

/*****************************************
 * Driver functions
 *****************************************/

void mgaTexEnv( GLcontext *ctx, GLenum pname, const GLfloat *param )
{
  if (!VALID_MGA_CONTEXT(mgaCtx)) {
    return;
  }

  if (pname == GL_TEXTURE_ENV_MODE) {
    mgaCtx->TextureMode = (int) *param;
    ctx->NewState |= NEW_DRVSTATE0;
  }
}

void mgaTexImage( GLcontext *ctx, GLenum target,
		  struct gl_texture_object *tObj, GLint level,
		  GLint internalFormat,
		  const struct gl_texture_image *image )
{
  mgaTextureObjectPtr t;

  CHECK_CONTEXT( return; );
  
  mgaMsg(10,"mgaTexImage():\n");
  if (target != GL_TEXTURE_2D)
    return;
  t = (mgaTextureObjectPtr) tObj->DriverData;
  if (!t) {
    t = mgaCreateTexObj(mgaCtx,tObj);
    tObj->DriverData = (void *) t;
    mgaMsg(10,"  new TexObj\n");
  } else
    if (!VALID_MGA_TEXTURE_OBJECT(t)) {
      mgaError("  invalid TexObj\n");
      return;
    }
  mgaMsg(10,"  Size: %d,%d; Level %d\n",image->Width,image->Height,
	 level);
	 
  if (CopyImage(t,level,internalFormat,image)) {
    mgaError("mgaTexImage: CopyImage failed!\n");  
  } 
     
  ctx->NewState |= NEW_DRVSTATE0;
}

void mgaTexSubImage( GLcontext *ctx, GLenum target,
		  struct gl_texture_object *tObj, GLint level,
		  GLint xoffset, GLint yoffset,
		  GLsizei width, GLsizei height,
		  GLint internalFormat,
		  const struct gl_texture_image *image )
{
  mgaMsg(10,"mgaTexSubImage():\n");
  mgaMsg(10,"  Size: %d,%d; Level %d\n",image->Width,image->Height,
           level);
}

void mgaTexParameter( GLcontext *ctx, GLenum target,
		      struct gl_texture_object *tObj,
		      GLenum pname, const GLfloat *params )
{
  mgaTextureObjectPtr t = (mgaTextureObjectPtr) tObj->DriverData;

  CHECK_CONTEXT( return; );

  if (!t || target != GL_TEXTURE_2D) {
    mgaMsg(1,"mgaTexParameter(): TexObj == 0 || target != GL_TEXTURE_2D\n");
    return;
  }

  switch (pname) {

  case GL_TEXTURE_MIN_FILTER:
  case GL_TEXTURE_MAG_FILTER:
    mgaSetTexFilter(t,tObj->MinFilter,tObj->MagFilter);
    break;

  case GL_TEXTURE_WRAP_S:
  case GL_TEXTURE_WRAP_T:
    mgaSetTexWrapping(t,tObj->WrapS,tObj->WrapT);
    break;
  
  case GL_TEXTURE_BORDER_COLOR:
    mgaSetTexBorderColor(t,tObj->BorderColor);
    break;

  default:
    return;
  }
  ctx->NewState |= NEW_DRVSTATE0;
}

void mgaBindTexture( GLcontext *ctx, GLenum target,
		     struct gl_texture_object *tObj )
{
  mgaTextureObjectPtr t;

  CHECK_CONTEXT( return; );

  if (target != GL_TEXTURE_2D)
    return;

  ctx->NewState |= NEW_DRVSTATE0;
  t = (mgaTextureObjectPtr) tObj->DriverData;
  if (!t) {
    t = mgaCreateTexObj(mgaCtx,tObj);
    tObj->DriverData = (void *) t;
    mgaMsg(10,"mgaBindTexture(): new TexObj\n");
  } else {
    if (!VALID_MGA_TEXTURE_OBJECT(t)) {
      mgaError("mgaBindTexture():  invalid TexObj\n");
      return;      
    }
   if ((t->swapped_out) && (mgaSwapInTexObj(t))) {
      mgaError("mgaBindTexture():  swap in failed\n");
      return;
    }
  }
  mgaCtx->CurrentTexObj = t;
  t->age = ++mgaCtx->current_texture_age;
}

void mgaDeleteTexture( GLcontext *ctx, struct gl_texture_object *tObj )
{
  mgaTextureObjectPtr t = (mgaTextureObjectPtr)tObj->DriverData;

  CHECK_CONTEXT( return; );
  mgaMsg(10,"mgaDeleteTexture():\n");
  if (t == mgaCtx->CurrentTexObj) {
    mgaMsg(10,"  delete current TexObj\n");
    mgaCtx->CurrentTexObj = mgaCtx->DefaultTexObj;
  }
  mgaDestroyTexObj(mgaCtx,t);
  tObj->DriverData=NULL;
}

void mgaUpdateTexturePalette( GLcontext *ctx, struct gl_texture_object *tObj )
{
  mgaTextureObjectPtr t;
  mgaUI16 *dst;
  mgaUI8  *src;
  int i, size, next;
  GLenum format;

  CHECK_CONTEXT( return; );
  
  if (tObj) {
    mgaMsg(8,"mgaUpdateTexturePalette(): size %d\n", tObj->PaletteSize);
    t = (mgaTextureObjectPtr) tObj->DriverData;;
    if (!t) return;
    size  = tObj->PaletteSize;
    format= tObj->PaletteFormat;
    dst   = t->Palette;
    src   = (mgaUI8 *) tObj->Palette;
  }
  else  { /* shared palette */
    mgaCtx->GlobalPaletteUpdated = 1;
    size  = ctx->Texture.PaletteSize;
    format= ctx->Texture.PaletteFormat;
    dst   = mgaCtx->GlobalPalette;
    src   = (mgaUI8 *) ctx->Texture.Palette;
    
    mgaMsg(8, "mgaUpdateTexturePalette(): size %d intFormat %x format %x\n",
	   size, ctx->Texture.PaletteIntFormat, ctx->Texture.PaletteFormat);
  }
  if (size > 256) {
    mgaError("mgaUpdateTexturePalette(): palette > 256 entries!\n");
    return;
  }
  switch (format)
  { 
    case GL_RGB: next = 3; break;
    case GL_RGBA:next = 4; break;
    default : 
      mgaError("mgaUpdateTexturePalette(): yet unsupported palette format %x\n");
      next = 4;
  }

  for (i=0; i < size; i++) {
    *dst = MGAPACKCOLOR565(src[RCOMP],src[GCOMP],src[BCOMP]);
    dst++;
    src += next;
  }
  
  
}
